import pydantic.v1.dataclasses
from pydantic import Field
from pydantic.v1 import validator
from tickit.adapters.io import TcpIo
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import ComponentConfig
from tickit.core.components.system_component import SystemComponent
from tickit.core.typedefs import ComponentID, ComponentPort, PortID

from tickit_devices.zebra._common import param_types
from tickit_devices.zebra.and_or_block import AndOrBlockConfig
from tickit_devices.zebra.zebra import ZebraAdapter


def _default() -> dict[str, int]:
    return dict.fromkeys(param_types.keys(), 0)


@pydantic.v1.dataclasses.dataclass
class Zebra(ComponentConfig):
    """
    Simulation of a Zebra device with a TCP server for reading/setting params/muxes
    (see `ZebraAdapter` for what read/set is available); Block wiring is currently
    invariant while Tickit is running, and only those blocks that are configured in
    components are instantiated.

    Configuration that is currently passed down to Block behaviour from `params`:
    - For AND/OR gates N=1,2,3,4, the following (default 0) may be set
    to Σ2**M where M is each input (1,2,3,4) for which the behaviour is desired.
    {AND|OR}{N}_INV: Inverts input(s) M
    {AND|OR}{N}_ENA: Enables input(s) M
    """

    name: ComponentID
    inputs: dict[PortID, ComponentPort]
    expose: dict[PortID, ComponentPort]
    components: list[AndOrBlockConfig]
    host: str = "localhost"
    port: int = 7012
    params: dict[str, int] = Field(default_factory=dict)

    @validator("params")
    def add_defaults(cls, v: dict[str, int]) -> dict[str, int]:  # noqa: N805
        return {**_default(), **v}

    def __call__(self) -> SystemComponent:
        return SystemComponent(
            adapter=AdapterContainer(
                adapter=ZebraAdapter(params=self.params),
                io=TcpIo(host=self.host, port=self.port),
            ),
            components=self.components,
            expose=self.expose,
            name=self.name,
        )



import pydantic.v1.dataclasses
from tickit.adapters.io import EpicsIo
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent

from .current import CurrentDevice
from .femto import FemtoAdapter, FemtoDevice


@pydantic.v1.dataclasses.dataclass
class Femto(ComponentConfig):
    """Femto simulation with EPICS IOC."""

    initial_gain: float = 2.5
    initial_current: float = 0.0
    db_file: str = "src/tickit_devices/femto/record.db"
    ioc_name: str = "FEMTO"

    def __call__(self) -> Component:  # noqa: D102
        device = FemtoDevice(
            initial_gain=self.initial_gain, initial_current=self.initial_current
        )
        adapters = [
            AdapterContainer(
                FemtoAdapter(device),
                EpicsIo(
                    self.ioc_name,
                    self.db_file,
                ),
            )
        ]
        return DeviceComponent(
            name=self.name,
            device=device,
            adapters=adapters,
        )


@pydantic.v1.dataclasses.dataclass
class Current(ComponentConfig):
    """Simulated current source."""

    callback_period: int = int(1e9)

    def __call__(self) -> Component:  # noqa: D102
        return DeviceComponent(
            name=self.name,
            device=CurrentDevice(callback_period=self.callback_period),
        )



import subprocess
import sys

from tickit_devices import __version__


def test_cli_version():
    cmd = [sys.executable, "-m", "tickit_devices", "--version"]
    assert subprocess.check_output(cmd).decode().strip() == __version__



from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path


@dataclass
class Image:
    """Dataclass to create a basic Image object."""

    index: int
    hash: str
    dtype: str
    data: bytes
    encoding: str
    shape: tuple[int, int]

    @classmethod
    def create_dummy_image(cls, index: int, shape: tuple[int, int]) -> "Image":
        """Returns an Image object wrapping the dummy blob using the metadata provided.

        Args:
            index (int): The index of the Image in the current acquisition.

        Returns:
            Image: An Image object wrapping the dummy blob.
        """
        data = dummy_image_blob()
        hsh = str(hash(data))
        dtype = "uint16"
        encoding = "bs16-lz4<"
        return Image(index, hsh, dtype, data, encoding, shape)


DUMMY_IMAGE_BLOB_PATH: Path = Path(__file__).parent / "frame_sample"


@lru_cache(maxsize=1)
def dummy_image_blob() -> bytes:
    """Load and cache the dummy image blob.

    Load the raw bytes of a compressed image
    taken from the stream of a real Eiger detector.
    This function is cached so there should be few
    (ideally one) loads per runtime.

    Returns:
        A compressed image as a bytes object.
    """
    with DUMMY_IMAGE_BLOB_PATH.open("rb") as frame_file:
        return frame_file.read()



from typing import TypedDict

from softioc import builder
from tickit.adapters.epics import EpicsAdapter
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime


class FemtoDevice(Device):
    """Electronic signal amplifier."""

    class Inputs(TypedDict):
        input: float

    class Outputs(TypedDict):
        current: float

    def __init__(
        self,
        initial_gain: float,
        initial_current: float,
    ) -> None:
        """Initialise the Femto device class.

        Args:
            initial_gain (Optional[float]): The initial amplified difference between \
                the input and output signals. Defaults to 2.5.
            initial_current (Optional[float]): The input signal current. \
                Defaults to 0.0.
        """
        self.gain = initial_gain
        self._current = initial_current
        self._output_current = initial_current * initial_gain

    def set_gain(self, gain: float) -> None:
        """Sets a new amplified difference between the input and output signals.

        Args:
            gain (float): The amplified difference between the input and output signals.
        """
        self.gain = gain

    def get_gain(self) -> float:
        """Returns the current amplified difference between the input and output
        signals.

        Returns:
            float: the amplified difference between the input and output signals.
        """
        return self.gain

    def set_current(self, input_current: float) -> None:
        """Set the output current based on the input current and the gain.

        The output current is calculated as:
            output_current = input_current * gain

        Args:
            input_current (float): The current of the input signal.
        """
        self._output_current = input_current * self.gain

    def get_current(self) -> float:
        """Returns the output current of the signal.

        Returns:
            float: The output current of the signal.
        """
        return self._output_current

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """Updates the state of the Femto device.

        Args:
            time (SimTime): The time of the simulation in nanoseconds.
            inputs (dict): The dict containing the input values of gain and current.

        Returns:
            DeviceUpdate: A container for the Device's outputs and a callback time.
        """
        current_value = inputs["input"]
        if current_value is not None:
            self.set_current(current_value)

        return DeviceUpdate(self.Outputs(current=self.get_current()), None)


class FemtoAdapter(EpicsAdapter):
    """The adapter for the Femto device."""

    device: FemtoDevice

    def __init__(self, device: FemtoDevice) -> None:
        super().__init__()
        self.device = device

    async def callback(self, value) -> None:
        """Device callback function.

        Args:
            value (float): The value to set the gain to.
        """
        self.device.set_gain(value)
        await self.interrupt()

    def on_db_load(self) -> None:
        """Customises records that have been loaded in to suit the simulation."""
        builder.aOut(
            "GAIN", initial_value=self.device.get_gain(), on_update=self.callback
        )
        self.link_input_on_interrupt(builder.aIn("GAIN_RBV"), self.device.get_gain)
        self.link_input_on_interrupt(builder.aIn("CURRENT"), self.device.get_current)



"""Eiger_status temporary docstring - to be changed."""

from dataclasses import dataclass, field, fields
from datetime import datetime
from enum import Enum
from typing import Any

from .eiger_schema import ro_float, ro_str


class State(Enum):
    """Possible states of the Eiger detector."""

    NA = "na"
    IDLE = "idle"
    READY = "ready"
    ACQUIRE = "acquire"
    CONFIGURE = "configure"
    INITIALIZE = "initialize"
    ERROR = "error"
    # TEST = "test"


def status_keys() -> list[str]:
    return [
        # "error",  # Eiger does not report error as a key
        "humidity",
        "link_0",
        "link_1",
        "series_unique_id",
        "state",
        "temperature",
        "time",
    ]


@dataclass
class EigerStatus:
    """Stores the status parameters of the Eiger detector."""

    state: State = field(
        default=State.NA,
        metadata=ro_str(allowed_values=[state.value for state in State]),
    )
    error: list[str] = field(default_factory=list, metadata=ro_str())
    temperature: float = field(default=24.5, metadata=ro_float())
    humidity: float = field(default=0.2, metadata=ro_float())
    time: datetime = field(default=datetime.now(), metadata=ro_str())
    dcu_buffer_free: float = field(default=0.5, metadata=ro_float())
    link_0: str = field(default="up", metadata=ro_str(allowed_values=["up", "down"]))
    link_1: str = field(default="up", metadata=ro_str(allowed_values=["up", "down"]))
    series_unique_id: str = field(
        default="01HBV3JPF9T4ZDPADX6EMK6XMZ", metadata=ro_str()
    )

    keys: list[str] = field(default_factory=status_keys)

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")



from dataclasses import dataclass, field, fields
from typing import Any

from tickit_devices.eiger.eiger_schema import rw_bool, rw_str, rw_uint


def monitor_config_keys() -> list[str]:
    return ["buffer_size", "discard_new", "mode"]


@dataclass
class MonitorConfig:
    """Eiger monitor configuration taken from the API spec."""

    mode: str = field(
        default="enabled", metadata=rw_str(allowed_values=["enabled", "disabled"])
    )
    buffer_size: int = field(default=512, metadata=rw_uint())
    discard_new: bool = field(default=False, metadata=rw_bool())

    keys: list[str] = field(default_factory=monitor_config_keys)

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        for field_ in fields(self):
            if field_.name == key:
                return {"value": vars(self)[field_.name], "metadata": field_.metadata}
        raise ValueError(f"No field with name {key}")

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: D105
        self.__dict__[key] = value



import pytest
from tickit.core.device import DeviceUpdate
from tickit.core.typedefs import SimTime

from tickit_devices.synchrotron.synchrotron_machine import (
    SynchrotronMachineStatusDevice,
)

MODE = 1
USERCOUNTDN = 200000
BEAMENERGY = 3.2


@pytest.fixture
def synchrotron_machine() -> SynchrotronMachineStatusDevice:
    return SynchrotronMachineStatusDevice(
        initial_mode=MODE, initial_countdown=USERCOUNTDN, initial_energy=BEAMENERGY
    )


def test_synchrotron_machine_constructor(
    synchrotron_machine: SynchrotronMachineStatusDevice,
):
    pass


def test_synchrotron_machine_set_and_get_mode(
    synchrotron_machine: SynchrotronMachineStatusDevice,
):
    synchrotron_machine.synchrotron_mode = MODE
    assert synchrotron_machine.get_mode() == MODE


def test_synchrotron_machine_update(
    synchrotron_machine: SynchrotronMachineStatusDevice,
):
    device_update: DeviceUpdate = synchrotron_machine.update(SimTime(0), inputs={})
    assert device_update.outputs["mode"] == 1
    assert device_update.call_at is None



import pathlib
from typing import TypedDict

import pydantic.v1.dataclasses
from softioc import builder
from tickit.adapters.epics import EpicsAdapter
from tickit.adapters.io import EpicsIo, TcpIo
from tickit.adapters.specifications import RegexCommand
from tickit.adapters.tcp import CommandAdapter
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.byte_format import ByteFormat


class SynchrotronCurrentDevice(Device):
    """Device to simulate the ring current signal value from the synchrotron.

    This SynchrotronCurrentDevice simulates only a single signal that can be read,
    the beam current.The real world pv for this current is SR-DI-DCCT-01:SIGNAL.

    The signal is read via and epics adapter, and set using a tcp adapter.
    """

    #: An empty typed mapping of device inputs
    class Inputs(TypedDict): ...

    #: A typed mapping containing the current output value
    class Outputs(TypedDict):
        current: float

    def __init__(
        self,
        initial_current: float | None,
        callback_period: int = int(1e9),
        countdown: float = 600.0,
        fill_time: float = 15.0,
        target_current: float = 300.0,
        minimum_current: float = 270.0,
    ) -> None:
        """Initialise the SynchrotonCurrent device.

        Args:
            initial_current (Optional[float]): The inital beam current. Defaults to
                300mA.
            callback_period: (int): The number of nanoseconds it will wait
                between calls
            countdown (float): Length of time in seconds to deplete
                charge from target_current to minimum_current.
            fill_time (float): Length of time in seconds to increase charge
                charge from target_current to minimum_current.
            target_current (float): The current the synchrotron should be topped up to.
            minimum_current (float): The current the synchrotron can fall to before
                being topped up.

        """
        self.target_current = target_current
        self.minimum_current = minimum_current
        self.beam_current = initial_current if initial_current else self.target_current
        self.callback_period = callback_period

        self.topup_fill = False

        # it should take 600 seconds to go from target_current 270, 15 seconds to fill
        self.loss_increment = (self.minimum_current - self.target_current) / countdown
        self.fill_increment = (self.target_current - self.minimum_current) / fill_time

        self.last_update_time = None

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """Update method that just outputs beam current.

        The device is only altered by adapters so take no inputs.
        The current is lost at a rate of ~0.02mA per second, during the topup
        phase it gains ~2mA per second

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the value of the beam current.
        """
        # check to see if topup fill should be activated/deactivated
        if self.topup_fill:
            self.topup_fill = self.beam_current < self.target_current
        else:
            self.topup_fill = self.beam_current <= self.minimum_current

        period = self.callback_period * 1e-9
        if self.last_update_time:
            period = time - self.last_update_time

        self.beam_current += (
            self.topup_fill * self.fill_increment  # if we're refilling
            + (not self.topup_fill) * self.loss_increment  # if we're not refilling
        ) * float(period)

        self.last_time = time
        call_at = SimTime(time + self.callback_period)
        return DeviceUpdate(
            SynchrotronCurrentDevice.Outputs(current=self.beam_current), call_at
        )

    def get_current(self) -> float:
        """Beam current getter for the epics adapter."""
        return self.beam_current


class SynchrotronCurrentTCPAdapter(CommandAdapter):
    """A TCP adapter to set a SynchrotronCurrentDevice PV values."""

    device: SynchrotronCurrentDevice
    _byte_format: ByteFormat = ByteFormat(b"%b\r\n")

    def __init__(self, device: SynchrotronCurrentDevice) -> None:
        super().__init__()
        self.device = device

    @RegexCommand(r"C=(\d+\.?\d*)", interrupt=True, format="utf-8")
    async def set_beam_current(self, value: float) -> None:
        """Regex string command that sets the value of beam_current.

        Args:
            value (int): The new value of beam_current.
        """
        self.device.beam_current = value

    @RegexCommand(r"C\?", format="utf-8")
    async def get_beam_current(self) -> bytes:
        """Regex string command that returns the utf-8 encoded value of beam_current.

        Returns:
            bytes: The utf-8 encoded value of beam_current.
        """
        return str(self.device.beam_current).encode("utf-8")


class SynchrotronCurrentEpicsAdapter(EpicsAdapter):
    """Epics adapter for reading device current as a PV through channel access."""

    device: SynchrotronCurrentDevice

    def __init__(self, device: SynchrotronCurrentDevice) -> None:
        super().__init__()
        self.device = device

    def on_db_load(self) -> None:
        """Link loaded in record with getter for device."""
        self.link_input_on_interrupt(builder.aIn("SIGNAL"), self.device.get_current)


@pydantic.v1.dataclasses.dataclass
class SynchrotronCurrent(ComponentConfig):
    """Synchrotron current component."""

    initial_current: float | None
    callback_period: int = int(1e9)
    host: str = "localhost"
    port: int = 25565
    db_file: str = str(pathlib.Path(__file__).parent.absolute() / "db_files/DCCT.db")
    ioc_name: str = "SR-DI-DCCT-01"

    def __call__(self) -> Component:  # noqa: D102
        device = SynchrotronCurrentDevice(
            self.initial_current,
            callback_period=self.callback_period,
        )
        adapters = [
            AdapterContainer(
                SynchrotronCurrentTCPAdapter(device),
                TcpIo(
                    self.host,
                    self.port,
                ),
            ),
            AdapterContainer(
                SynchrotronCurrentEpicsAdapter(device),
                EpicsIo(
                    self.ioc_name,
                    self.db_file,
                ),
            ),
        ]
        return DeviceComponent(
            name=self.name,
            device=device,
            adapters=adapters,
        )



import pytest

from tickit_devices.eiger.filewriter.filewriter_config import FileWriterConfig

# # # # # Eiger FileWriterConfig Tests # # # # #


@pytest.fixture
def filewriter_config() -> FileWriterConfig:
    return FileWriterConfig()


def test_eiger_filewriter_config_constructor():
    FileWriterConfig()


def test_eiger_filewriter_config_get_set(filewriter_config):
    assert "enabled" == filewriter_config["mode"]["value"]
    filewriter_config["mode"] = "disabled"
    assert "disabled" == filewriter_config["mode"]["value"]

    with pytest.raises(ValueError):
        filewriter_config["doesnt_exist"]



import pytest
from tickit.core.typedefs import ComponentID, SimTime

from tickit_devices.zebra.and_or_block import AndOrBlock

all_true = 15
one_false = 14
one_true = 1
all_false = 0
varying_values_of_true = {all_false, one_true, one_false, all_true}


# # # # # AndOrBlock Tests # # # # #


@pytest.mark.parametrize("enabled", varying_values_of_true)
@pytest.mark.parametrize("inverted", varying_values_of_true)
def test_and_block_enabled_or_inverted(enabled: int, inverted: int):
    name = ComponentID("AND1")
    mux = {f"INP{i + 1}": True for i in range(4)}
    block = AndOrBlock(name=name)
    block.params = {f"{name}_ENA": enabled, f"{name}_INV": inverted}
    out = block.update(SimTime(0), mux)
    assert out.outputs == {
        "OUT": enabled + inverted == all_true
    }  # All devices either enabled or inverted


@pytest.mark.parametrize("enabled", varying_values_of_true)
@pytest.mark.parametrize("inverted", varying_values_of_true)
def test_or_block_enabled_or_inverted(enabled: int, inverted: int):
    name = ComponentID("OR1")
    block = AndOrBlock(name=name)
    block.params = {f"{name}_ENA": enabled, f"{name}_INV": inverted}
    mux = {f"INP{i + 1}": True for i in range(4)}
    out = block.update(SimTime(0), mux)
    assert out.outputs == {
        "OUT": enabled != inverted
    }  # At least one device exclusively enabled or inverted


@pytest.mark.parametrize("high", varying_values_of_true)
def test_all_enabled_and_block_for_inputs(high: int):
    name = ComponentID("AND1")
    block = AndOrBlock(name=name)
    block.params = {f"{name}_ENA": all_true, f"{name}_INV": all_false}
    mux = {f"INP{i + 1}": bool(high & (1 << i)) for i in range(4)}
    out = block.update(SimTime(0), mux)
    assert out.outputs == {"OUT": high == all_true}  # All inputs high


@pytest.mark.parametrize("high", varying_values_of_true)
def test_all_enabled_or_block_for_inputs(high: int):
    name = ComponentID("OR1")
    block = AndOrBlock(name=name)
    block.params = {f"{name}_ENA": all_true, f"{name}_INV": all_false}
    mux = {f"INP{i + 1}": bool(high & (1 << i)) for i in range(4)}
    out = block.update(SimTime(0), mux)
    assert out.outputs == {"OUT": high != all_false}  # At least one input high



import datetime
from typing import Any

import cbor2
import pytest

from tickit_devices.eiger.data.dummy_image import Image
from tickit_devices.eiger.eiger import EigerDevice
from tickit_devices.eiger.eiger_settings import EigerSettings
from tickit_devices.eiger.stream.eiger_stream_2 import EigerStream2


@pytest.fixture
def stream() -> EigerStream2:
    return EigerStream2()


@pytest.fixture
def eiger() -> EigerDevice:
    return EigerDevice()


TEST_SERIES_ID = 15614

EIGER_SETTINGS_HEADER = EigerSettings().filtered(
    ["flatfield", "pixelmask", "countrate_correction_table"]
)
X_SIZE = EIGER_SETTINGS_HEADER["x_pixels_in_detector"]
Y_SIZE = EIGER_SETTINGS_HEADER["y_pixels_in_detector"]

ALL_FIELDS = ["flatfield", "pixel_mask", "countrate_correction_lookup_table"]

BASIC_START_MESSAGE: dict[str, Any] = {
    "type": "start",
    "series_id": 15614,
    "series_unique_id": "01HBV3JPF9T4ZDPADX6EMK6XMZ",
    "arm_date": datetime.datetime(
        2023,
        10,
        3,
        17,
        47,
        48,
        329000,
        tzinfo=datetime.timezone(datetime.timedelta(seconds=7200)),
    ),
    "beam_center_x": 2049.3840906675064,
    "beam_center_y": 2163.621048575148,
    "channels": ["threshold_1"],
    "count_time": 0.004317472232502031,
    "countrate_correction_lookup_table": b"\x00\x00\x00\x00",
    "countrate_correction_enabled": True,
    "detector_description": "Dectris EIGER2 Si 16M",
    "detector_serial_number": "E-32-0117",
    "detector_translation": [
        0.15370380680006296,
        0.1622715786431361,
        -0.23715919962215962,
    ],
    "flatfield": {"threshold_1": b"\x00\x00\x00\x00"},
    "flatfield_enabled": True,
    "frame_time": 0.004317572232502031,
    "goniometer": {
        "chi": {"increment": 0.0, "start": 30.0},
        "omega": {"increment": 0.1, "start": 0.0},
        "phi": {"increment": 0.0, "start": 0.0},
    },
    "image_size_x": 4148,
    "image_size_y": 4362,
    "incident_energy": 13500.299829398293,
    "incident_wavelength": 0.918381073013,
    "number_of_images": 1,
    "pixel_mask": {"threshold_1": b"\x00\x00\x00\x00"},
    "pixel_mask_enabled": True,
    "pixel_size_x": 7.5e-05,
    "pixel_size_y": 7.5e-05,
    "saturation_value": 21517,
    "sensor_material": "Si",
    "sensor_thickness": 0.00045,
    "threshold_energy": {
        "threshold_1": 6750.149914699146,
        "threshold_2": 18900.41976115761,
    },
    "virtual_pixel_interpolation_enabled": True,
}


# This does not include 'image_id' or 'data' as these are tested separately
IMAGE_MESSAGE = {
    "type": "image",
    "series_id": 15614,
    "series_unique_id": "01HBV3JPF9T4ZDPADX6EMK6XMZ",
    "real_time": [215873, 50000000],
    "series_date": datetime.datetime(
        2023,
        10,
        3,
        17,
        47,
        49,
        434000,
        tzinfo=datetime.timezone(datetime.timedelta(seconds=7200)),
    ),
    "start_time": [0, 50000000],
    "stop_time": [215873, 50000000],
}


END_MESSAGE = {
    "type": "end",
    "series_id": 15614,
    "series_unique_id": "01HBV3JPF9T4ZDPADX6EMK6XMZ",
}


@pytest.mark.parametrize(
    "header_detail",
    [("none"), ("basic")],
)
def test_begin_series_message_basic(stream: EigerStream2, header_detail: str) -> None:
    settings = EigerSettings()

    stream.begin_series(settings, TEST_SERIES_ID, header_detail)
    data = list(stream.consume_data())[0]
    assert isinstance(data, bytes)
    message = cbor2.loads(data)
    for axis, start_value in BASIC_START_MESSAGE["goniometer"].items():
        assert start_value == message["goniometer"][axis]
    message.pop("goniometer")
    reduced_start_message = BASIC_START_MESSAGE.copy()
    reduced_start_message.pop("goniometer")
    for f in ALL_FIELDS:
        reduced_start_message.pop(f)
    assert message == reduced_start_message
    assert not any(f in message for f in ALL_FIELDS)


def test_begin_series_message_all(stream: EigerStream2) -> None:
    settings = EigerSettings()
    stream.begin_series(settings, TEST_SERIES_ID, "all")

    message = cbor2.loads(list(stream.consume_data())[0])

    for axis, start_value in BASIC_START_MESSAGE["goniometer"].items():
        assert start_value == message["goniometer"][axis]

    message.pop("goniometer")
    reduced_start_message = BASIC_START_MESSAGE.copy()
    reduced_start_message.pop("goniometer")
    # the ALL_FIELDS entries get set to default value, overwriting initial message
    for f in ALL_FIELDS:
        assert f in message
        message.pop(f)
        reduced_start_message.pop(f)

    assert message == reduced_start_message


def test_insert_image_produces_correct_message(stream: EigerStream2) -> None:
    for i in range(2):
        image = Image.create_dummy_image(i, (X_SIZE, Y_SIZE))

        stream.insert_image(image, TEST_SERIES_ID)
        message = cbor2.loads(list(stream.consume_data())[0])

        # Image data is too big to compare - just sanity check size
        assert message["data"]["threshold_1"].value[0] == [Y_SIZE, X_SIZE]
        del message["data"]

        # Check image_id and remove it to compare against generic expected message
        assert message["image_id"] == i
        del message["image_id"]

        assert message == IMAGE_MESSAGE


def test_end_series_produces_correct_message(stream: EigerStream2) -> None:
    stream.end_series(TEST_SERIES_ID)

    message = stream.consume_data()
    message = cbor2.loads(list(stream.consume_data())[0])

    assert message == END_MESSAGE


def test_data_buffered(stream: EigerStream2) -> None:
    settings = EigerSettings()
    image = Image.create_dummy_image(0, (X_SIZE, Y_SIZE))

    stream.begin_series(settings, TEST_SERIES_ID, "basic")
    stream.insert_image(image, TEST_SERIES_ID)
    stream.insert_image(image, TEST_SERIES_ID)
    stream.insert_image(image, TEST_SERIES_ID)
    stream.end_series(TEST_SERIES_ID)

    messages = [cbor2.loads(b) for b in stream.consume_data()]

    assert [m["type"] for m in messages] == ["start", "image", "image", "image", "end"]



import logging
from typing import TypedDict

import pydantic.v1.dataclasses
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime

LOGGER = logging.getLogger(__name__)


@pydantic.v1.dataclasses.dataclass
class Counter(ComponentConfig):
    """Simulation of simple counting device."""

    def __call__(self) -> Component:  # noqa: D102
        return DeviceComponent(name=self.name, device=CounterDevice())


class CounterDevice(Device):
    """A simple device which increments a value."""

    #: An empty typed mapping of input values
    class Inputs(TypedDict): ...

    #: A typed mapping containing the 'value' output value
    class Outputs(TypedDict):
        value: int

    def __init__(self, initial_value: int = 0, callback_period: int = int(1e9)) -> None:
        """A constructor of the counter, which increments the input value.

        Args:
            initial_value (Any): The initial value of the counter.
            callback_period (int): The simulation time callback period of the device
                (in nanoseconds). Defaults to int(1e9).
        """
        self._value = initial_value
        self.callback_period = SimTime(callback_period)
        LOGGER.debug(f"Counter initialized with value => {self._value}")

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which produces the incremented value.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the incremented value, and
                requests a callback after callback_period.
        """
        self._value = self._value + 1
        LOGGER.debug(f"Counter incremented to {self._value}")
        return DeviceUpdate(
            CounterDevice.Outputs(value=self._value),
            SimTime(time + self.callback_period),
        )


class DividingDevice(Device):
    """
    A toy device which turns an input number into a boolean of whether it
    divides exactly into a given denominator.
    """

    class Inputs(TypedDict):
        input: int

    class Outputs(TypedDict):
        output: bool

    def __init__(self, denominator: int) -> None:
        self.denominator = denominator

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        output = inputs["input"] % self.denominator == 0
        return DeviceUpdate(self.Outputs(output=output), None)


@pydantic.v1.dataclasses.dataclass
class DividingConfig(ComponentConfig):
    denominator: int

    def __call__(self) -> DeviceComponent:
        """Create the component from the given config."""
        return DeviceComponent(
            name=self.name, device=DividingDevice(denominator=self.denominator)
        )


class FizzBangDevice(Device):
    class Inputs(TypedDict):
        fizzbang: bool
        fizz: bool
        bang: bool

    class Outputs(TypedDict):
        output: str

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        out = (
            ["fizzbang"]
            if inputs["fizzbang"]
            else [k for k, v in inputs.items() if v] or ["Not Fizzbang"]
        )
        return DeviceUpdate(self.Outputs(output=out[0]), None)


@pydantic.v1.dataclasses.dataclass
class FizzBang(ComponentConfig):
    def __call__(self) -> DeviceComponent:
        """Create the component from the given config."""
        return DeviceComponent(name=self.name, device=FizzBangDevice())



import aiohttp
import pytest

DETECTOR_URL = "http://localhost:8081/detector/api/1.8.0/"
FILE_WRITER_URL = "http://localhost:8081/filewriter/api/1.8.0/"
MONITOR_URL = "http://localhost:8081/monitor/api/1.8.0/"
STREAM_URL = "http://localhost:8081/stream/api/1.8.0/"
REQUEST_TIMEOUT = 1.0


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tickit_task", ["examples/configs/eiger/eiger.yaml"], indirect=True
)
async def test_eiger_system(tickit_task):
    commands = {
        "initialize": {"sequence id": 1},
        "disarm": {"sequence id": 3},
        "cancel": {"sequence id": 5},
        "abort": {"sequence id": 6},
    }

    headers = {"content-type": "application/json"}

    async def get_status(status, expected):
        async with session.get(
            DETECTOR_URL + f"status/{status}",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert expected == (await response.json())["value"]

    async with aiohttp.ClientSession() as session:
        await get_status(status="state", expected="na")

        # Test each command
        for key, value in commands.items():
            async with session.put(
                DETECTOR_URL + f"command/{key}",
                timeout=REQUEST_TIMEOUT,
            ) as response:
                assert value == (await response.json())

        # Check status
        await get_status(status="board_000/th0_temp", expected=24.5)
        await get_status(status="board_000/th0_humidity", expected=0.2)
        await get_status(status="builder/dcu_buffer_free", expected=0.5)

        # Test Eiger in IDLE state
        await get_status(status="state", expected="idle")

        async with session.get(
            STREAM_URL + "status/keys",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert (await response.json()) == ["dropped", "state"]

        # Test settings/getting config
        async with session.get(
            DETECTOR_URL + "config/element",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert (await response.json())["value"] == "Co"

        async with session.put(
            DETECTOR_URL + "config/element",
            headers=headers,
            json={"value": "Li"},
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert (await response.json()) == ["element"]

        async with session.get(
            DETECTOR_URL + "config/photon_energy",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert 54.3 == (await response.json())["value"]

        async with session.get(
            FILE_WRITER_URL + "config/mode",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert "enabled" == (await response.json())["value"]

        async with session.put(
            FILE_WRITER_URL + "config/mode",
            headers=headers,
            json={"value": "enabled"},
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert [] == (await response.json())

        # Test filewriter, monitor and stream endpoints
        async with session.get(
            FILE_WRITER_URL + "status/state",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert "ready" == (await response.json())["value"]

        async with session.get(
            MONITOR_URL + "config/mode",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert "enabled" == (await response.json())["value"]

        async with session.put(
            MONITOR_URL + "config/mode",
            headers=headers,
            json={"value": "enabled"},
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert [] == (await response.json())

        async with session.get(
            MONITOR_URL + "status/error",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert [] == (await response.json())["value"]

        async with session.get(
            STREAM_URL + "config/mode",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert "enabled" == (await response.json())["value"]

        async with session.put(
            STREAM_URL + "config/mode",
            headers=headers,
            json={"value": "enabled"},
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert [] == (await response.json())

        async with session.put(
            DETECTOR_URL + "config/threshold/1/energy",
            headers=headers,
            json={"value": 6829},
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert [
                "flatfield",
                "threshold/1/energy",
                "threshold/1/flatfield",
                "threshold/2/flatfield",
                "threshold_energy",
            ] == (await response.json())

        async with session.get(
            DETECTOR_URL + "config/threshold/1/energy",
            headers=headers,
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert 6829 == (await response.json())["value"]

        async with session.put(
            DETECTOR_URL + "config/threshold/difference/mode",
            headers=headers,
            json={"value": "enabled"},
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert ["difference_mode"] == (await response.json())

        async with session.get(
            DETECTOR_URL + "config/threshold/difference/mode",
            headers=headers,
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert "enabled" == (await response.json())["value"]

        # Test acquisition in ints mode
        async with session.put(
            DETECTOR_URL + "config/trigger_mode",
            headers=headers,
            json={"value": "ints"},
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert ["trigger_mode"] == (await response.json())

        async with session.get(
            STREAM_URL + "status/state",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert "ready" == (await response.json())["value"]

        await get_status(status="state", expected="idle")

        async with session.put(
            DETECTOR_URL + "config/ntrigger",
            headers=headers,
            json={"value": 2},
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert ["ntrigger"] == (await response.json())

        async with session.put(
            DETECTOR_URL + "command/arm",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert {"sequence id": 2} == (await response.json())

        await get_status(status="state", expected="ready")

        async with session.put(
            DETECTOR_URL + "command/trigger",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert {"sequence id": 4} == (await response.json())

        await get_status(status="state", expected="ready")

        async with session.put(
            DETECTOR_URL + "command/trigger",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert {"sequence id": 4} == (await response.json())

        await get_status(status="state", expected="idle")

        # Test we get a 404 for a non-existent URI
        async with session.get(
            DETECTOR_URL + "status/doesnt_exist",
            timeout=REQUEST_TIMEOUT,
        ) as response:
            assert response.status == 404



import asyncio
import logging
from collections.abc import Mapping
from queue import Queue

from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from typing_extensions import TypedDict

from tickit_devices.eiger.data.dummy_image import Image
from tickit_devices.eiger.eiger_settings import EigerSettings
from tickit_devices.eiger.filewriter.filewriter_config import FileWriterConfig
from tickit_devices.eiger.filewriter.filewriter_status import FileWriterStatus
from tickit_devices.eiger.monitor.monitor_config import MonitorConfig
from tickit_devices.eiger.monitor.monitor_status import MonitorStatus
from tickit_devices.eiger.stream.eiger_stream import EigerStream
from tickit_devices.eiger.stream.eiger_stream_2 import EigerStream2
from tickit_devices.eiger.stream.stream_config import (
    CBOR_STREAM,
    LEGACY_STREAM,
    StreamConfig,
)
from tickit_devices.eiger.stream.stream_status import StreamStatus

from .eiger_status import EigerStatus, State

LOGGER = logging.getLogger("Eiger")


class EigerDevice(Device):
    """Simulation logic for the Eiger detector.

    The simulation acquires frames based on the commands called. It supports
    the following state transitions:

    NA -> IDLE
    IDLE -> READY
    READY -> IDLE
    READY -> ACQUIRING
    ACQUIRING -> READY
    ACQUIRING -> IDLE
    """

    settings: EigerSettings
    status: EigerStatus
    stream: EigerStream | EigerStream2
    streams: Mapping[str, EigerStream | EigerStream2]

    _num_frames_left: int
    _data_queue: Queue

    class Inputs(TypedDict, total=False):
        trigger: bool

    class Outputs(TypedDict): ...

    def __init__(
        self,
        settings: EigerSettings | None = None,
        status: EigerStatus | None = None,
        stream: EigerStream | EigerStream2 | None = None,
    ) -> None:
        """Construct a new eiger.

        Args:
            settings: Eiger settings. Defaults to None.
            status: Starting status. Defaults to None.
            stream: Data stream handler. Defaults to None.
        """
        self.settings = settings or EigerSettings()
        self.status = status or EigerStatus()

        self.stream_status: StreamStatus = StreamStatus()
        self.stream_config: StreamConfig = StreamConfig()
        self.streams = {
            LEGACY_STREAM: stream or EigerStream(callback_period=SimTime(int(1e9))),
            CBOR_STREAM: stream or EigerStream2(callback_period=SimTime(int(1e9))),
        }
        self.stream = self.streams[CBOR_STREAM]

        self.filewriter_status: FileWriterStatus = FileWriterStatus()
        self.filewriter_config: FileWriterConfig = FileWriterConfig()
        self.filewriter_callback_period = SimTime(int(1e9))

        self.monitor_status: MonitorStatus = MonitorStatus()
        self.monitor_config: MonitorConfig = MonitorConfig()
        self.monitor_callback_period = SimTime(int(1e9))

        self._num_frames_left: int = 0
        self._num_triggers_left: int = 0
        self._total_frames: int = 0
        self._data_queue: Queue = Queue()
        self._series_id: int = 0

        self._finished_trigger: asyncio.Event | None = None

    @property
    def finished_trigger(self) -> asyncio.Event:
        """Event that is set when an acqusition series is complete.

        Property ensures the event is created.
        """
        if self._finished_trigger is None:
            self._finished_trigger = asyncio.Event()

        return self._finished_trigger

    async def initialize(self) -> None:
        """Initialize the detector.

        Required for all subsequent operations.
        """
        self._set_state(State.IDLE)

    async def arm(self) -> None:
        """Arm the detector.

        Required for triggering.
        """
        self.stream = self.streams[self.stream_config.format]
        self._series_id += 1
        self.stream.begin_series(
            self.settings, self._series_id, self.stream_config.header_detail
        )
        self._num_frames_left = self.settings.nimages
        self._num_triggers_left = self.settings.ntrigger
        self._set_state(State.READY)

    async def disarm(self) -> None:
        """Disarm the detector.

        Intended for use when armed. See state diagram in class docstring.
        """
        self._set_state(State.IDLE)
        self.stream.end_series(self._series_id)

    async def trigger(self) -> None:
        """Trigger the detector.

        If the detector is in INTS mode, it will begin acquiring frames the
        next time update() is called. If it is in EXTS mode, this call will
        be ignored and acquisition will start based on the parameter to
        update().
        INTE and EXTE mode are currently not supported.
        """
        LOGGER.info("Trigger requested")
        trigger_mode = self.settings.trigger_mode

        if self._is_in_state(State.READY) and trigger_mode == "ints":
            self._begin_acqusition_mode()
        else:
            LOGGER.info(
                f"Ignoring trigger, state={self.get_state()},"
                f"trigger_mode={trigger_mode}"
            )

    async def cancel(self) -> None:
        """Cancel acquisition.

        The detector will stop acquiring frames after the next full frame is taken,
        it will then return to a READY state as though it has just been armed.
        """
        self._set_state(State.READY)
        self.stream.end_series(self._series_id)

    async def abort(self) -> None:
        """Abort acquisition.

        The detector will immediately stop acquiring frames and disarm itself.
        """
        self._set_state(State.IDLE)
        self.stream.end_series(self._series_id)

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """Update the detector.

        Depending on the detector's current state, will begin, continue or
        clean up an acquisition series.

        Args:
            time: The current simulation time (in nanoseconds).
            inputs: A mapping of device inputs and their values.
        """
        if self._is_in_state(State.ACQUIRE):
            if self._num_frames_left > 0:
                self._acquire_frame()

                return DeviceUpdate(
                    self.Outputs(), SimTime(time + int(self.settings.frame_time * 1e9))
                )
            else:
                self.finished_trigger.set()

                if self._num_triggers_left > 0:
                    self._set_state(State.READY)
                    self._num_frames_left = self.settings.nimages
                else:
                    LOGGER.debug("Ending Series...")
                    self._set_state(State.IDLE)
                    self.stream.end_series(self._series_id)

        if inputs.get("trigger", False):
            self._begin_acqusition_mode()
            # Should have another update immediately to begin acquisition
            return DeviceUpdate(self.Outputs(), SimTime(time))

        return DeviceUpdate(self.Outputs(), None)

    def _begin_acqusition_mode(self) -> None:
        self._num_triggers_left -= 1
        self._set_state(State.ACQUIRE)
        LOGGER.info("Now in acquiring mode")
        self.finished_trigger.clear()

    def _acquire_frame(self) -> None:
        frame_id = (
            (self.settings.ntrigger - self._num_triggers_left) * self.settings.nimages
        ) - self._num_frames_left
        LOGGER.debug(f"Frame id {frame_id}")

        shape = (
            self.settings.x_pixels_in_detector,
            self.settings.y_pixels_in_detector,
        )
        image = Image.create_dummy_image(frame_id, shape)
        self.stream.insert_image(image, self._series_id)
        self._num_frames_left -= 1
        LOGGER.debug(f"Frames left: {self._num_frames_left}")
        LOGGER.debug(f"Triggers left: {self._num_triggers_left}")

    def get_state(self) -> State:
        """Get the eiger's current state

        Returns:
            State: The state the detector is in.
                See state diagram in class docstring.
        """
        return self.status.state

    def _set_state(self, state: State) -> None:
        self.status.state = state

    def _is_in_state(self, state: State) -> bool:
        return self.get_state() is state


def get_changed_parameters(key: str) -> list[str]:
    """Get the list of parameters that may have changed as a result of putting
    to the parameter provided.

    Args:
        key: string key of the changed parameter within the detector subsystem

    Returns:
        list[str]: a list of keys which may have been changed after a PUT request
    """
    match key:
        case "auto_summation":
            return ["auto_summation", "frame_count_time"]
        case "count_time" | "frame_time":
            return [
                "bit_depth_image",
                "bit_depth_readout",
                "count_time",
                "countrate_correction_count_cutoff",
                "frame_count_time",
                "frame_time",
            ]
        case "flatfield":
            return ["flatfield", "threshold/1/flatfield"]
        case "incident_energy" | "photon_energy":
            return [
                "element",
                "flatfield",
                "incident_energy",
                "photon_energy",
                "threshold/1/energy",
                "threshold/1/flatfield",
                "threshold/2/energy",
                "threshold/2/flatfield",
                "threshold_energy",
                "wavelength",
            ]
        case "pixel_mask":
            return ["pixel_mask", "threshold/1/pixel_mask"]
        case "threshold/1/flatfield":
            return ["flatfield", "threshold/1/flatfield"]
        case "roi_mode":
            return ["count_time", "frame_time", "roi_mode"]
        case "threshold_energy" | "threshold/1/energy":
            return [
                "flatfield",
                "threshold/1/energy",
                "threshold/1/flatfield",
                "threshold/2/flatfield",
                "threshold_energy",
            ]
        case "threshold/2/energy":
            return [
                "flatfield",
                "threshold/1/flatfield",
                "threshold/2/energy",
                "threshold/2/flatfield",
            ]
        case "threshold/1/mode":
            return ["threshold/1/mode", "threshold/difference/mode"]
        case "threshold/2/mode":
            return ["threshold/2/mode", "threshold/difference/mode"]
        case "threshold/1/pixel_mask":
            return ["pixel_mask", "threshold/1/pixel_mask"]
        case "threshold/difference/mode":
            return ["difference_mode"]  # replicating API inconsistency
        case _:
            return [key]



import pydantic.v1.dataclasses
from tickit.adapters.io import TcpIo
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent

from .cryostream import CryostreamAdapter, CryostreamDevice


@pydantic.v1.dataclasses.dataclass
class Cryostream(ComponentConfig):
    """Cryostream simulation with TCP server."""

    host: str = "localhost"
    port: int = 25565

    def __call__(self) -> Component:  # noqa: D102
        device = CryostreamDevice()
        adapters = [
            AdapterContainer(
                CryostreamAdapter(device),
                TcpIo(
                    self.host,
                    self.port,
                ),
            )
        ]
        return DeviceComponent(
            name=self.name,
            device=device,
            adapters=adapters,
        )



from enum import IntEnum


class RunModes(IntEnum):
    """An enumerator for Cryostream run modes."""

    STARTUP = 0
    STARTUPFAIL = 1
    STARTUPOK = 2
    RUN = 3
    SETUP = 4
    SHUTDOWNOK = 5
    SHUTDOWNFAIL = 6


class PhaseIds(IntEnum):
    """An enumerator for Cryostream phases."""

    RAMP = 0
    COOL = 1
    PLAT = 2
    HOLD = 3
    END = 4
    PURGE = 5


class AlarmCodes(IntEnum):
    """An enumerator for Cryostream alarm codes alarm levels."""

    NO_ERRORS = 0
    STOP_PRESSED = 1
    STOP_COMMAND = 2
    END_COMPLETE = 3
    PURGE_COMPLETE = 4
    TEMP_WARNING = 5
    PRESSURE_WARNING = 6
    CHECK_VACUUM = 7
    SELF_CHECK_FAIL = 8
    FLOW_RATE = 9
    TEMP_CONTROL_ERROR = 10
    GAS_TYPE_ERROR = 11
    TEMP_READING_ERROR = 12
    SUCT_TEMP_ERROR = 13
    SENSOR_FAIL = 14
    BROWNOUT = 15
    SINK_OVERHEAT = 16
    PSU_OVERHEAT = 17
    POWER_LOSS = 18
    COLHEAD_TOO_COLD = 19
    COLHEAD_TIMEOUT = 20
    CRYODRIVE_NOT_FOUND = 21
    CRYODRIVE_ERROR = 22
    NO_NITROGEN = 23
    NO_HELIUM = 24
    VAC_GAUGE_FAIL = 25
    VAC_READING_ERROR = 26
    RS232_ERROR = 27
    COLDHEAD_TEMP_WARNING = 28
    COLDHEAD_TEMP_ERROR = 29
    DO_NOT_OPEN_CRYOSTAT_L2 = 30
    DO_NOT_OPEN_CRYOSTAT_L3 = 31
    UNPLUG_XTAL_SENSOR = 32
    CRYOSTAT_OPEN = 33
    CRYOSTAT_OPEN_TIMEOUT = 34
    HIGH_TEMP_WARNING = 35
    HIGH_TEMP_ERROR = 36
    CRYODRIVE_T_SENSOR_FAULT = 37
    CRYODRIVE_P_SENSOR_FAULT = 38
    CRYODRIVE_LOW_T_TRIP = 39
    CRYODRIVE_HIGH_T_TRIP = 40
    CRYODRIVE_LOW_P_TRIP = 41
    CRYODRIVE_HIGH_T_WARNING = 42
    CRYODRIVE_LOW_P_WARNING = 43
    CONNECT_GAS_SUPPLY = 44
    AUTOFILL_FAULT = 45
    AUTOFILL_ABOUT_TO_FILL = 46
    AUTOFILL_FILLING = 47
    COLLAR_TEMP_ERROR = 48
    COLDHEAD_ERROR = 49
    TURBO_FLOW = 50
    HE_SELECTED = 51
    CRYODRIVE_NOT_READY = 52
    REGEN_REQUIRED = 53
    REGEN_COMPLETE = 54
    CONNECT_VACUUM = 55
    DISCONNECT_VACUUM = 56


class HardwareType(IntEnum):
    """An enumerator for Cryostream hardware types."""

    PLUS_SYSTEM = 1  # Max temp 500 K (Cryostream, Cobra and Smartstream)
    HAS_CRYOSHUTTER = 2  # 700 series Cryostream and Cobra
    SERIES_800 = 3
    HAS_AUTOFILL = 4  # 800 series Cryostream from firmware version 150



import pytest

from tickit_devices.eiger.monitor.monitor_status import MonitorStatus

# # # # # Eiger MonitorStatus Tests # # # # #


@pytest.fixture
def monitor_status() -> MonitorStatus:
    return MonitorStatus()


def test_eiger_monitor_status_constructor():
    MonitorStatus()


def test_eiger_monitor_status_getitem(monitor_status):
    assert [] == monitor_status["error"]["value"]

    with pytest.raises(ValueError):
        monitor_status["doesnt_exist"]



import asyncio
import struct
from collections.abc import AsyncIterable
from typing import TypedDict

from tickit.adapters.specifications import RegexCommand
from tickit.adapters.tcp import CommandAdapter
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime

from tickit_devices.cryostream.base import CryostreamBase
from tickit_devices.cryostream.states import PhaseIds

_EXTENDED_STATUS = ">BBHHHBBHHHHHBBBBBBHHBBBBBBBBHH"


class CryostreamDevice(Device, CryostreamBase):
    """A Cryostream device, used for cooling of samples using cold gas."""

    #: An empty typed mapping of device inputs
    class Inputs(TypedDict): ...

    #: A typed mapping containing the 'temperature' output value
    class Outputs(TypedDict):
        temperature: float

    def __init__(self) -> None:
        """A Cryostream constructor sets up initial internal values."""
        super().__init__()
        self.phase_id: int = PhaseIds.HOLD.value
        self.callback_period: SimTime = SimTime(int(1e9))

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which changes the output temperature according to set
        modes.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the value of the output
                temperature, and requests callback if temperature should continue to
                change.
        """
        if self.phase_id in (PhaseIds.RAMP.value, PhaseIds.COOL.value):
            self.gas_temp = self.update_temperature(time)
            return DeviceUpdate(
                self.Outputs(temperature=self.gas_temp),
                SimTime(time + self.callback_period),
            )
        if self.phase_id == PhaseIds.PLAT.value:
            self.phase_id = PhaseIds.HOLD.value
            return DeviceUpdate(
                self.Outputs(temperature=self.gas_temp),
                SimTime(time + int(self.plat_duration * 1e10)),
            )
        return DeviceUpdate(self.Outputs(temperature=self.gas_temp), None)


class CryostreamAdapter(CommandAdapter):
    """A Cryostream TCP adapter which sends regular status packets and can set modes."""

    device: CryostreamDevice

    def __init__(self, device: CryostreamDevice) -> None:
        super().__init__()
        self.device = device

    async def on_connect(self) -> AsyncIterable[bytes]:
        """A method which continiously yields status packets.

        Returns:
            AsyncIterable[bytes]: An asyncronous iterable of packed Cryostream status
                packets.
        """
        while True:
            await asyncio.sleep(2.0)
            await self.device.set_status_format(1)
            status = await self.device.get_status(1)
            yield status.pack()

    @RegexCommand(b"\\x02\\x0a", interrupt=True)
    async def restart(self) -> None:
        """A regex bytes command which restarts the Cryostream."""
        await self.device.restart()

    @RegexCommand(b"\\x02\\x0d", interrupt=True)
    async def hold(self) -> None:
        """A regex bytes command which holds the current temperature."""
        await self.device.hold()

    @RegexCommand(b"\\x02\\x10", interrupt=True)
    async def purge(self) -> None:
        """A regex bytes command which purges (immediately raise to 300K)."""
        await self.device.purge()

    @RegexCommand(b"\\x02\\x11", interrupt=True)
    async def pause(self) -> None:
        """A regex bytes command which pauses."""
        await self.device.pause()

    @RegexCommand(b"\\x02\\x12", interrupt=True)
    async def resume(self) -> None:
        """A regex bytes command which resumes the last command."""
        await self.device.resume()

    @RegexCommand(b"\\x02\\x13", interrupt=True)
    async def stop(self) -> None:
        """A regex bytes command which stops gas flow."""
        await self.device.stop()

    @RegexCommand(b"\\x03\\x14([\\x00\\x01])", interrupt=True)
    async def turbo(self, turbo_on: bytes) -> None:
        """A regex bytes command which enables / disables turbo mode.

        Args:
            turbo_on (bytes): The desired turbo mode, where 0 denotes off and 1 denotes
                on.
        """
        turbo_on = struct.unpack(">B", turbo_on)[0]
        await self.device.turbo(turbo_on)  # type: ignore

    # Todo set status format not interrupt
    @RegexCommand(b"\\x03\\x28([\\x00\\x01])", interrupt=False)
    async def set_status_format(self, status_format: bytes) -> None:
        """A regex bytes command which sets the status packet format.

        Args:
            status_format (bytes): The status packet format, where 0 denotes a standard
                status packet and 1 denotes an extended status packet.
        """
        status_format = struct.unpack(">B", status_format)[0]
        await self.device.set_status_format(status_format)  # type: ignore

    @RegexCommand(b"\\x04\\x0c(.{2})", interrupt=True)
    async def plat(self, duration: bytes) -> None:
        """A regex bytes command which maintains temperature for a set amount of time.

        Args:
            duration (bytes): The duration for which the temperature should be held.
        """
        duration = struct.unpack(">H", duration)[0]
        await self.device.plat(duration)  # type: ignore

    @RegexCommand(b"\\x04\\x0f(.{2})", interrupt=True)
    async def end(self, ramp_rate: bytes) -> None:
        """A regex bytes command which brings the gas temperature to 300 K at ramp rate.

        Args:
            ramp_rate (bytes): The rate at which the temperature should change.
        """
        ramp_rate = struct.unpack(">H", ramp_rate)[0]
        await self.device.end(ramp_rate)  # type: ignore

    @RegexCommand(b"\\x04\\x0e(.{2})", interrupt=True)
    async def cool(self, target_temp: bytes) -> None:
        """A regex bytes command which makes decreases temperature to a set value.

        Args:
            target_temp (bytes): The target temperature.
        """
        target_temp = struct.unpack(">H", target_temp)[0]
        await self.device.cool(target_temp)  # type: ignore

    @RegexCommand(b"\\x06\\x0b(.{2,4})", interrupt=True)
    async def ramp(self, values: bytes) -> None:
        """Change gas temperature to a set value at a controlled rate.

        Args:
            values (bytes): The rate at which the temperature should change and the
                target temperature.
        """
        ramp_rate, target_temp = struct.unpack(">HH", values)
        await self.device.ramp(ramp_rate, target_temp)  # type: ignore



import pathlib
from typing import TypedDict

import pydantic.v1.dataclasses
from softioc import builder
from tickit.adapters.epics import EpicsAdapter
from tickit.adapters.io import EpicsIo, TcpIo
from tickit.adapters.specifications import RegexCommand
from tickit.adapters.tcp import CommandAdapter
from tickit.core.adapter import AdapterContainer
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_component import DeviceComponent
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.byte_format import ByteFormat


class SynchrotronMachineStatusDevice(Device):
    """Device to simulate the machine status records from the synchrotron.

    The signal is read via an epics adapter, and set using a tcp adapter.
    """

    class Inputs(TypedDict): ...

    class Outputs(TypedDict):
        mode: int
        user_countdown: float
        beam_energy: float

    def __init__(
        self,
        initial_mode: int,
        initial_countdown: float,
        initial_energy: float,
    ) -> None:
        """Initialise the SynchrotonMachineStatus device.

        Args:
            initial_mode (int): The inital synchrotron operation status, defaults at 0,
                or "Shutdown".
            initial_countdown (float): Initial Time countdown to the beam dump
            initial_energy (float): Inital current value of ring energy
        """
        self.synchrotron_mode = initial_mode
        self.user_countdown = initial_countdown
        self.beam_energy = initial_energy

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """Update method that outputs machine status record values.

        The device is only altered by adapters so take no inputs and is independent of
        time.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the value of the machine
                status records.
        """
        return DeviceUpdate(
            SynchrotronMachineStatusDevice.Outputs(
                mode=self.synchrotron_mode,
                user_countdown=self.user_countdown,
                beam_energy=self.beam_energy,
            ),
            None,
        )

    def get_mode(self) -> int:
        """Return value of synchrotron mode, required for epics adapter."""
        return self.synchrotron_mode

    def get_user_countdown(self) -> float:
        """Return value of user countdown, required for epics adapter."""
        return self.user_countdown

    def get_beam_energy(self) -> float:
        """Return value of beam energy, required for epics adapter."""
        return self.beam_energy


class SynchrotronMachineStatusTCPAdapter(CommandAdapter):
    """A TCP adapter to set a SynchrotronMachineStatusDevice PV values."""

    device: SynchrotronMachineStatusDevice
    _byte_format: ByteFormat = ByteFormat(b"%b\r\n")

    def __init__(self, device: SynchrotronMachineStatusDevice) -> None:
        super().__init__()
        self.device = device

    @RegexCommand(r"mode=([0-7])", interrupt=True, format="utf-8")
    async def set_synchrotron_mode(self, value: int) -> None:
        """Regex string command to set the value of synchrotron_mode.

        Args:
            value (int): The new value of synchrotron mode in digits 0-7.
                These numbers are used by the mbbi record, which returns
                the corresponding status word on a caget.
        """
        self.device.synchrotron_mode = value

    @RegexCommand(r"UCD=(\d+\.?\d*)", interrupt=True, format="utf-8")
    async def set_user_countdown(self, value: float) -> None:
        """Regex string command to set the value of user_countdown.

        Args:
            value (int): The new value of user_countdown.
        """
        self.device.user_countdown = value

    @RegexCommand(r"BE=(\d+\.?\d*)", interrupt=True, format="utf-8")
    async def set_beam_energy(self, value: float) -> None:
        """Regex string command to set the value of beam_energy.

        Args:
            value (int): The new value of beam_energy.
        """
        self.device.beam_energy = value

    @RegexCommand(r"mode\?", format="utf-8")
    async def get_synchrotron_mode(self) -> bytes:
        """Regex string command that returns synchrotron_mode.

        Returns:
            bytes: The utf-8 encoded value of synchrotron_mode.
        """
        return str(self.device.synchrotron_mode).encode("utf-8")

    @RegexCommand(r"UCD\?", format="utf-8")
    async def get_user_countdown(self) -> bytes:
        """Regex string command that returns user_countdown.

        Returns:
            bytes: The utf-8 encoded value of user_countdown.
        """
        return str(self.device.user_countdown).encode("utf-8")

    @RegexCommand(r"BE\?", format="utf-8")
    async def get_beam_energy(self) -> bytes:
        """Regex string command that returns beam_energy.

        Returns:
            bytes: The utf-8 encoded value of beam_energy.
        """
        return str(self.device.beam_energy).encode("utf-8")


class SynchrotronMachineStatusEpicsAdapter(EpicsAdapter):
    """Epics adapter for reading device values as a PV through channel access."""

    device: SynchrotronMachineStatusDevice

    def __init__(self, device: SynchrotronMachineStatusDevice) -> None:
        super().__init__()
        self.device = device

    def on_db_load(self) -> None:
        """Link epics records with getters for device.

        The MODE record is loaded via db, the other two are created here.
        """
        self.link_input_on_interrupt(builder.mbbIn("MODE"), self.device.get_mode)
        self.link_input_on_interrupt(
            builder.aIn("USERCOUNTDN"),
            self.device.get_user_countdown,
        )
        self.link_input_on_interrupt(
            builder.aIn("BEAMENERGY"), self.device.get_beam_energy
        )


@pydantic.v1.dataclasses.dataclass
class SynchrotronMachineStatus(ComponentConfig):
    """Synchrotron Machine status component."""

    initial_mode: int = 4
    initial_countdown: float = 100000
    initial_energy: float = 3.0
    host: str = "localhost"
    port: int = 25565
    db_file: str = str(pathlib.Path(__file__).parent.absolute() / "db_files/MSTAT.db")
    ioc_name: str = "CS-CS-MSTAT-01"

    def __call__(self) -> Component:  # noqa: D102
        device = SynchrotronMachineStatusDevice(
            self.initial_mode,
            self.initial_countdown,
            self.initial_energy,
        )
        adapters = [
            AdapterContainer(
                SynchrotronMachineStatusTCPAdapter(device),
                TcpIo(
                    self.host,
                    self.port,
                ),
            ),
            AdapterContainer(
                SynchrotronMachineStatusEpicsAdapter(device),
                EpicsIo(
                    self.ioc_name,
                    self.db_file,
                ),
            ),
        ]
        return DeviceComponent(
            name=self.name,
            device=device,
            adapters=adapters,
        )
